@@ -36,14 +36,7 @@ from rdiffweb.core.i18n import ugettext as _
 from rdiffweb.core.ldap_auth import LdapPasswordStore
 from rdiffweb.core.librdiff import RdiffRepo, DoesNotExistError, \
     AccessDeniedError
-from rdiffweb.core.store_sqlite import SQLiteBackend
-
-
-try:
-    # Python 2.5+
-    from hashlib import sha1 as sha
-except ImportError:
-    from sha import new as sha
+from rdiffweb.core.passwd import check_password, hash_password
 
 # Define the logger
 logger = logging.getLogger(__name__)
@@ -54,6 +47,12 @@ MAX_DEPTH = 5
 
 DEFAULT_REPO_ENCODING = codecs.lookup((sys.getfilesystemencoding() or 'utf-8').lower()).name
 
+# Define roles
+ADMIN_ROLE = 0
+MAINTAINER_ROLE = 5
+USER_ROLE = 10
+ROLES = [ADMIN_ROLE, MAINTAINER_ROLE, USER_ROLE]
+
 
 def normpath(val):
     """
@@ -84,16 +83,11 @@ def _split_path(path):
         return path.decode('utf-8'), b''
 
 
-def _hash_password(password):
-    # At this point the password should be unicode. We converted it into
-    # system encoding.
-    password_b = password.encode('utf8')
-    hasher = sha()
-    hasher.update(password_b)
-    value = hasher.hexdigest()
-    if isinstance(value, bytes):
-        value = value.decode(encoding='latin1')
-    return value
+class DuplicateSSHKeyError(Exception):
+    """
+    Raised by add_authorizedkey when trying to add the same SSH Key twice.
+    """
+    pass
 
 
 class IUserChangeListener():
@@ -169,8 +163,8 @@ class UserObject(object):
         # Parse and validate ssh key
         assert key
         key = authorizedkeys.check_publickey(key)
-        if not key:
-            raise ValueError(_("Invalid SSH key."))
+        assert key, "invalid ssh key"
+
         # Remove option, replace comments.
         key = authorizedkeys.AuthorizedKey(
             options=None,
@@ -183,7 +177,7 @@ class UserObject(object):
         if os.path.isfile(filename):
             with open(filename, mode="r+", encoding='utf-8') as fh:
                 if authorizedkeys.exists(fh, key):
-                    raise ValueError(_("SSH key already exists"))
+                    raise DuplicateSSHKeyError(_("SSH key already exists"))
                 logger.info("add key [%s] to [%s] authorized_keys", key, self.username)
                 authorizedkeys.add(fh, key)
         else:
@@ -196,7 +190,7 @@ class UserObject(object):
                     key=key.getvalue())
                 assert inserted
             except sqlite3.IntegrityError:  # @UndefinedVariable
-                raise ValueError(_("Duplicate key. This key already exists or is associated to another user."))
+                raise DuplicateSSHKeyError(_("Duplicate key. This key already exists or is associated to another user."))
         self._store._notify('user_attr_changed', self, {'authorizedkeys': True })
 
     def add_repo(self, repopath):
@@ -290,7 +284,7 @@ class UserObject(object):
 
     def _get_attr(self, key):
         """Return user's attribute"""
-        assert key in ['userid', 'username', 'isadmin', 'useremail', 'userroot', 'password'], "invalid attribute: " + key
+        assert key in ['userid', 'username', 'role', 'useremail', 'userroot', 'password'], "invalid attribute: " + key
         if not self._record:
             self._record = self._db.findone('users', userid=self._userid)
         return self._record[key]
@@ -337,7 +331,7 @@ class UserObject(object):
                or self._db.findone('repos', userid=self.userid, repopath=repopath + "/")
                or self._db.findone('repos', userid=self.userid, repopath="/" + repopath + "/"))
         if not row:
-            raise DoesNotExistError(repopath)
+            raise DoesNotExistError(self.userid, repopath)
         return RepoObject(self, row)
 
     def _get_repos(self):
@@ -352,9 +346,16 @@ class UserObject(object):
         """Return True if this user is an LDAP user. (with a password)"""
         return not self._get_attr('password')
 
+    def _is_role(self, role):
+        assert role in ROLES
+        try:
+            return int(self._get_attr('role')) <= role
+        except:
+            return False
+
     def _set_attr(self, obj_key, key, value, notify=True):
         """Used to define an attribute"""
-        assert key in ['isadmin', 'useremail', 'userroot', 'password'], "invalid attribute: " + key
+        assert key in ['role', 'useremail', 'userroot', 'password'], "invalid attribute: " + key
         updated = self._db.update('users', userid=self._userid, **{key: value})
         assert updated, 'update failed'
         if self._record:
@@ -391,9 +392,9 @@ class UserObject(object):
             except:
                 pass
         # Fallback to database
-        if old_password and self._get_attr('password') != _hash_password(old_password):
+        if old_password and not check_password(old_password, self.hash_password):
             raise ValueError(_("Wrong password"))
-        self._set_attr('password', 'password', _hash_password(password))
+        self.hash_password = hash_password(password)
         self._store._notify('user_password_changed', self.username, password)
 
     def update_repos(self):
@@ -409,6 +410,8 @@ class UserObject(object):
             for name in dirs:
                 if name == b'rdiff-backup-data':
                     repopath = os.path.relpath(root, start=user_root)
+                    # Handle special scenario when the repo is the user_root
+                    repopath = b'' if repopath == b'.' else repopath
                     try:
                         self.get_repo(repopath)
                     except DoesNotExistError:
@@ -418,14 +421,17 @@ class UserObject(object):
 
     # Declare properties
     userid = property(fget=lambda x: x._get_attr('userid'))
-    is_admin = property(fget=lambda x: x._get_attr('isadmin'), fset=lambda x, y: x._set_attr('is_admin', 'isadmin', y))
+    is_admin = property(fget=lambda x: x._is_role(ADMIN_ROLE))
+    is_maintainer = property(fget=lambda x: x._is_role(MAINTAINER_ROLE))
     email = property(fget=lambda x: x._get_attr('useremail'), fset=lambda x, y: x._set_attr('email', 'useremail', y))
     user_root = property(fget=lambda x: x._get_attr('userroot'), fset=lambda x, y: x._set_attr('user_root', 'userroot', y))
     username = property(fget=lambda x: x._get_attr('username'))
     repos = property(fget=lambda x: list(map(lambda y: y.strip('/'), x._get_repos())))
+    role = property(fget=lambda x: x._get_attr('role'), fset=lambda x, y: x._set_attr('role', 'role', int(y)))
     authorizedkeys = property(fget=lambda x: x._get_authorizedkeys())
     repo_objs = property(fget=lambda x: [RepoObject(x, r) for r in x._get_repos()])
     disk_quota = property(_get_disk_quota, _set_disk_quota)
+    hash_password = property(fget=lambda x: x._get_attr('password'), fset=lambda x, y: x._set_attr('password', 'password', y))
 
 
 @python_2_unicode_compatible
@@ -542,6 +548,7 @@ class Store():
 
     def __init__(self, app):
         self.app = app
+        from rdiffweb.core.store_sqlite import SQLiteBackend
         self._database = SQLiteBackend(self._db_file)
         self._password_stores = [LdapPasswordStore(app)]
         self._change_listeners = []
@@ -561,7 +568,7 @@ class Store():
         # Check if admin user exists. If not, created it.
         if not self.get_user(self._admin_user):
             userobj = self.add_user(self._admin_user, 'admin123')
-            userobj.is_admin = True
+            userobj.role = ADMIN_ROLE
 
     def add_change_listener(self, listener):
         self._change_listeners.append(listener)
@@ -581,7 +588,7 @@ class Store():
         # Find a database where to add the user
         logger.debug("adding new user [%s]", user)
         if password:
-            inserted = self._database.insert('users', username=user, password=_hash_password(password))
+            inserted = self._database.insert('users', username=user, password=hash_password(password))
         else:
             inserted = self._database.insert('users', username=user, password='')
         assert inserted
@@ -661,7 +668,7 @@ class Store():
             users = self._database.search('users', search, 'username', 'useremail')
         elif criteria:
             if criteria == 'admins':
-                users = self._database.find('users', isadmin=1)
+                users = self._database.find('users', role=ADMIN_ROLE)
             elif criteria == 'ldap':
                 users = self._database.find('users', password='')
             else:
@@ -706,35 +713,27 @@ class Store():
         assert password is None or isinstance(user, str)
         # Validate credential using database first.
         logger.debug("validating user [%s] credentials", user)
-        record = self._database.findone('users', username=user)
-        if record and record['password'] == _hash_password(password):
-            userobj = UserObject(self, record)
+        userobj = self.get_user(user)
+        if userobj and userobj.hash_password:
+            if not check_password(password, userobj.hash_password):
+                return None
             self._notify('user_logined', userobj, None)
             return userobj
 
         # Fallback to LDAP
-        for store in self._password_stores:
-            try:
-                valid = store.are_valid_credentials(user, password)
-                if valid:
-                    break
-            except:
-                pass
-        if not valid:
-            return None
-        # Get real username and external attributes.
-        real_user, attrs = valid
-        # Check if user exists in database
-        userobj = self.get_user(real_user)
-        if not userobj:
-            if self._allow_add_user:
-                # Create user
-                userobj = self.add_user(real_user, attrs=attrs)
-            else:
-                logger.info("user [%s] not found in database", real_user)
-                return None
-        self._notify('user_logined', userobj, attrs)
-        return userobj
+        if userobj or self._allow_add_user:
+            for store in self._password_stores:
+                try:
+                    valid = store.are_valid_credentials(user, password)
+                    if valid:
+                        real_user, attrs = valid
+                        if not userobj:
+                            userobj = self.add_user(real_user, attrs=attrs)
+                        self._notify('user_logined', userobj, attrs)
+                        return userobj
+                except:
+                    logger.warn('fail to validate credentials', exc_info=1)
+        return None
 
     def _notify(self, mod, *args):
         for listener in self._change_listeners:
