@@ -46,7 +46,7 @@ class MockRdiffRepo(RdiffRepo):
 
     def __init__(self):
         p = bytes(pkg_resources.resource_filename('rdiffweb.core', 'tests'), encoding='utf-8')  # @UndefinedVariable
-        RdiffRepo.__init__(self, os.path.dirname(p), os.path.basename(p))
+        RdiffRepo.__init__(self, os.path.dirname(p), os.path.basename(p), encoding='utf-8')
         self.root_path = MockDirEntry(self)
 
 
@@ -219,7 +219,7 @@ class RdiffRepoTest(unittest.TestCase):
         # Define location of testcases
         self.testcases_dir = os.path.normpath(os.path.join(self.temp_dir, 'testcases'))
         self.testcases_dir = self.testcases_dir.encode('utf8')
-        self.repo = RdiffRepo(self.temp_dir, b'testcases')
+        self.repo = RdiffRepo(self.temp_dir, b'testcases', encoding='utf-8')
 
     def tearDown(self):
         shutil.rmtree(self.temp_dir.encode('utf8'), True)
@@ -236,11 +236,11 @@ class RdiffRepoTest(unittest.TestCase):
         self.assertEqual('testcases', self.repo.display_name)
 
     def test_init_with_absolute(self):
-        self.repo = RdiffRepo(self.temp_dir, '/testcases')
+        self.repo = RdiffRepo(self.temp_dir, '/testcases', encoding='utf-8')
         self.assertEqual('testcases', self.repo.display_name)
 
     def test_init_with_invalid(self):
-        self.repo = RdiffRepo(self.temp_dir, 'invalid')
+        self.repo = RdiffRepo(self.temp_dir, 'invalid',encoding='utf-8')
         self.assertEqual('failed', self.repo.status[0])
         self.assertEqual(None, self.repo.last_backup_date)
         self.assertEqual(b'invalid', self.repo.path)
@@ -248,7 +248,7 @@ class RdiffRepoTest(unittest.TestCase):
 
     def test_get_path_root(self):
         dir_entry = self.repo.get_path(b"/")
-        self.assertEqual('', dir_entry.display_name)
+        self.assertEqual('testcases', dir_entry.display_name)
         self.assertEqual(b'', dir_entry.path)
         self.assertEqual(self.testcases_dir, dir_entry.full_path)
         self.assertEqual(True, dir_entry.isdir)
@@ -291,26 +291,19 @@ class RdiffRepoTest(unittest.TestCase):
         self.assertEqual('', status[1])
 
     def test_restore_file(self):
-        filename, stream = self.repo.restore(b"Revisions/Data", restore_date=1454448640, kind='zip')
-        self.assertEqual('Data', filename)
-        data = stream.read()
-        self.assertEqual(b'Version3\n', data)
-
-    def test_restore_direntry(self):
-        entry = self.repo.get_path(b"Revisions/Data")
-        filename, stream = self.repo.restore(entry, restore_date=1454448640, kind='zip')
+        filename, stream = self.repo.get_path(b"Revisions/Data").restore(restore_as_of=1454448640, kind='zip')
         self.assertEqual('Data', filename)
         data = stream.read()
         self.assertEqual(b'Version3\n', data)
 
     def test_restore_subdirectory(self):
-        filename, stream = self.repo.restore(b"Revisions/", restore_date=1454448640, kind='zip')
+        filename, stream = self.repo.get_path(b"Revisions/").restore(restore_as_of=1454448640, kind='zip')
         self.assertEqual('Revisions.zip', filename)
         data = stream.read()
         self.assertTrue(data)
 
     def test_restore_root(self):
-        filename, stream = self.repo.restore(b"/", restore_date=1454448640, kind='zip')
+        filename, stream = self.repo.get_path(b"/").restore(restore_as_of=1454448640, kind='zip')
         self.assertEqual('testcases.zip', filename)
         data = stream.read()
         self.assertTrue(data)
